/*
 * Copyright Elasticsearch B.V. and/or licensed to Elasticsearch B.V. under one
 * or more contributor license agreements. Licensed under the Elastic License
 * 2.0 and the Server Side Public License, v 1; you may not use this file except
 * in compliance with, at your election, the Elastic License 2.0 or the Server
 * Side Public License, v 1.
 */

/**
 * The vulnerability fields describe information about a vulnerability that is relevant to an event.
 */
export interface EcsVulnerability {
  /**
   * The type of system or architecture that the vulnerability affects. These may be platform-specific (for example, Debian or SUSE) or general (for example, Database or Firewall). For example (https://qualysguard.qualys.com/qwebhelp/fo_portal/knowledgebase/vulnerability_categories.htm[Qualys vulnerability categories])
   * This field must be an array.
   */
  category?: string[];
  /**
   * The classification of the vulnerability scoring system. For example (https://www.first.org/cvss/)
   */
  classification?: string;
  /**
   * The description of the vulnerability that provides additional context of the vulnerability. For example (https://cve.mitre.org/about/faqs.html#cve_entry_descriptions_created[Common Vulnerabilities and Exposure CVE description])
   */
  description?: string;
  /**
   * The type of identifier used for this vulnerability. For example (https://cve.mitre.org/about/)
   */
  enumeration?: string;
  /**
   * The identification (ID) is the number portion of a vulnerability entry. It includes a unique identification number for the vulnerability. For example (https://cve.mitre.org/about/faqs.html#what_is_cve_id)[Common Vulnerabilities and Exposure CVE ID]
   */
  id?: string;
  /**
   * A resource that provides additional information, context, and mitigations for the identified vulnerability.
   */
  reference?: string;
  /**
   * The report or scan identification number.
   */
  report_id?: string;
  scanner?: {
    /**
     * The name of the vulnerability scanner vendor.
     */
    vendor?: string;
  };

  score?: {
    /**
     * Scores can range from 0.0 to 10.0, with 10.0 being the most severe.
     * Base scores cover an assessment for exploitability metrics (attack vector, complexity, privileges, and user interaction), impact metrics (confidentiality, integrity, and availability), and scope. For example (https://www.first.org/cvss/specification-document)
     */
    base?: number;
    /**
     * Scores can range from 0.0 to 10.0, with 10.0 being the most severe.
     * Environmental scores cover an assessment for any modified Base metrics, confidentiality, integrity, and availability requirements. For example (https://www.first.org/cvss/specification-document)
     */
    environmental?: number;
    /**
     * Scores can range from 0.0 to 10.0, with 10.0 being the most severe.
     * Temporal scores cover an assessment for code maturity, remediation level, and confidence. For example (https://www.first.org/cvss/specification-document)
     */
    temporal?: number;
    /**
     * The National Vulnerability Database (NVD) provides qualitative severity rankings of "Low", "Medium", and "High" for CVSS v2.0 base score ranges in addition to the severity ratings for CVSS v3.0 as they are defined in the CVSS v3.0 specification.
     * CVSS is owned and managed by FIRST.Org, Inc. (FIRST), a US-based non-profit organization, whose mission is to help computer security incident response teams across the world. For example (https://nvd.nist.gov/vuln-metrics/cvss)
     */
    version?: string;
  };

  /**
   * The severity of the vulnerability can help with metrics and internal prioritization regarding remediation. For example (https://nvd.nist.gov/vuln-metrics/cvss)
   */
  severity?: string;
}
